#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

#include "modify_disk_image.h"


static char *make_new_filename(const char *filename);
static int read_file(const char *filename, unsigned char **data, size_t *size);
static int write_file(const char *filename, unsigned char *data, size_t size);
static int confirm(const char *msg);


int modify_disk_image(const char *filename, const struct patch_info_t *patch_info)
{
	int result = -1;
	unsigned char *data = NULL;
	size_t size = 0;
	int i;
	char *new_filename = NULL;
	struct stat st;

	printf("read original file: %s\n", filename);
	if (read_file(filename, &data, &size) < 0) {
		printf("ERROR: can't read file: %s\n", filename);
		goto ERROR_EXIT;
	}

	if (size != DISK_IMAGE_SIZE) {
		printf("ERROR: file size mismatch: %s size=%zu\n", filename, size);
		goto ERROR_EXIT;
	}

	printf("modify image\n");
	for (i = 0; 0 <= patch_info[i].offset; i++) {
		int offset = patch_info[i].offset;
		if (size <= offset) {
			printf("BUG: patch_info[%d] offset=%08X old=%02X new=%02X\n", i, offset, patch_info[i].old_value, patch_info[i].new_value);
			goto ERROR_EXIT;
		}
		if (data[offset] == patch_info[i].old_value) {
			data[offset] = patch_info[i].new_value;
		} else {
			printf("ERROR: mismatch: %s: offset=%08X value=%02X\n", filename, offset, data[offset]);
			goto ERROR_EXIT;
		}
	}

	if ((new_filename = make_new_filename(filename)) == NULL) {
		printf("ERROR: can't make output filename\n");
		goto ERROR_EXIT;
	}

	printf("write modified file: %s\n", new_filename);
	if (stat(new_filename, &st) == 0) {
		printf("already exists '%s'\n", new_filename);
		if (confirm("overwrite ?") <= 0) {
			goto ERROR_EXIT;
		}
	}

	if (write_file(new_filename, data, size) < 0) {
		printf("ERROR: can't write file: %s\n", new_filename);
		goto ERROR_EXIT;
	}

	result = 0;

ERROR_EXIT:
	if (new_filename != NULL) {
		free(new_filename);
	}
	if (data != NULL) {
		free(data);
	}
	return result;
}

static char *make_new_filename(const char *filename)
{
	char *newname = NULL;
	const char *ptr;
	int size;
	int baselen;

	ptr = strrchr(filename, '/');
#ifdef WIN32
	{
		const char *ptr2;

		if ((ptr2 = strrchr(filename, '\\')) != NULL) {
			if (ptr == NULL || ptr < ptr2) {
				ptr = ptr2;
			}
		}
	}
#endif
	if (ptr != NULL) {
		filename = ++ptr;
	}

	size = strlen(filename) + 4 + 1;

	if ((newname = (char *)malloc(size)) == NULL) {
		printf("ERROR: memory exhausted!\n");
		return NULL;
	}

	if ((ptr = strrchr(filename, '.')) == NULL) {
		return strcat(strcpy(newname, filename), ".NEW");
	}

	baselen = ptr - filename;
	strncpy(newname, filename, baselen);
	newname[baselen] = '\0';
	return strcat(strcat(newname, ".NEW"), ptr);
}

static int read_file(const char *filename, unsigned char **data, size_t *size)
{
	int result = -1;
	FILE *fp = NULL;

	*data = NULL;
	*size = 0;

	if ((fp = fopen(filename, "rb")) == NULL) {
		printf("ERROR: open error: %s\n", filename);
		goto ERROR_EXIT;
	}

	if (fseek(fp, 0, SEEK_END) < 0) {
		printf("ERROR: seek error: %s\n", filename);
		goto ERROR_EXIT;
	}

	*size = ftell(fp);

	if (fseek(fp, 0, SEEK_SET) < 0) {
		printf("ERROR: seek error: %s\n", filename);
		goto ERROR_EXIT;
	}

	if ((*data = (unsigned char *)malloc(*size)) == NULL) {
		printf("ERROR: memory exhausted!\n");
		goto ERROR_EXIT;
	}

	if (fread(*data, *size, 1, fp) != 1) {
		printf("ERROR: read error: %s\n", filename);
		goto ERROR_EXIT;
	}

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}
	return result;
}

static int write_file(const char *filename, unsigned char *data, size_t size)
{
	int result = -1;
	FILE *fp = NULL;

	if ((fp = fopen(filename, "wb")) == NULL) {
		printf("ERROR: open error: %s\n", filename);
		goto ERROR_EXIT;
	}

	if (fwrite(data, size, 1, fp) != 1) {
		printf("ERROR: write error: %s\n", filename);
		goto ERROR_EXIT;
	}

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}
	return result;
}

static int confirm(const char *msg)
{
	printf("%s (Y or N): ", msg);
	fflush(stdout);

	while (1) {
		int c = getchar();
		if (c == 'y' || c == 'Y') {
			return 1;
		}
		if (c == 'n' || c == 'N') {
			return 0;
		}
	}

	return -1;
}

/* EOF */
